from essentials import socket_ops_v2 as socket_ops
from .. import verification
import threading
from . import messaging
from .. import discovery
from essentials.network_ops import Get_IP_with_MAC
import time



class IOT_Device(object):

    def __init__(self, HOST=None, PORT=None, on_connected=None, on_data=None, on_question=None, on_video_frame=None, on_audio_frame=None, on_disconnect=None, receive_confirm=True, receive_confirm_timeout=5, password=None, challengeType=verification.types.NONCE, reconnect=True, meta=None):
        self.HOST = HOST
        self.PORT = PORT
        self.reconnect = reconnect
        self.connected = False
        self.password = password
        self.challengeType = challengeType
        self.verified_on_connection = on_connected
        self.config = socket_ops.Configuration()
        self.config.on_question = self.__questions__
        self.config.on_data_recv = self.__data__
        self.config.on_connection_close = self.__disconnected__
        self.config.server_PYTHONIC_only = True
        self.server = socket_ops.Socket_Connector(self.HOST, self.PORT, self.config)
        self.data_usage = self.server.data_usage
        self.ip_data = Get_IP_with_MAC()['ext'][0]
        self.__connect_lock__ = False

        self.on_data = on_data
        self.on_question = on_question
        self.on_video_frame = on_video_frame
        self.on_audio_frame = on_audio_frame
        self.on_disconnect = on_disconnect
        self.receive_confirm = receive_confirm
        self.receive_confirm_timeout = receive_confirm_timeout

        self.meta = meta

    def push_meta(self):
        msg = messaging.message(self.server, self.meta, self.receive_confirm, self.receive_confirm_timeout, messaging.types.META_PUSH, send_as_daemon=False)
        return msg

    def send_video_frame(self, frame, wait_till_confirm=False):
        msg = messaging.message(self.server, frame, self.receive_confirm, self.receive_confirm_timeout, messaging.types.VIDEO, send_as_daemon= not wait_till_confirm)
        return msg

    def send_audio_frame(self, frame, wait_till_confirm=False):
        msg = messaging.message(self.server, frame, self.receive_confirm, self.receive_confirm_timeout, messaging.types.AUDIO, send_as_daemon= not wait_till_confirm)
        return msg

    def send_data(self, data, wait_till_confirm=False):
        msg = messaging.message(self.server, data, self.receive_confirm, self.receive_confirm_timeout, messaging.types.DATA, send_as_daemon= not wait_till_confirm)
        return msg

    def ask(self, data, timeout=5):
        msg = messaging.message(self.server, data, True, timeout, messaging.types.QUESTION)
        return msg

    def ping(self, wait_till_confirm=True):
        msg = messaging.message(self.server, "PING", self.receive_confirm, self.receive_confirm_timeout, messaging.types.PING, send_as_daemon=not wait_till_confirm)
        return msg

    def __data__(self, data):
        if data == "error on verification":
            self.server.shutdown()
            raise PermissionError("Connecting Server Error: Invalid Credentials")

        if data == "__connect_lock__ disable":
            self.__connect_lock__ = False
            return
        
        message = messaging.translate_data(data, self)

        if "VIDEO" == message.message_type:
            if self.on_video_frame is not None:
                threading.Thread(target=self.on_video_frame, args=[message], daemon=True).start()
        elif "AUDIO" in message.message_type:
            if self.on_audio_frame is not None:
                threading.Thread(target=self.on_audio_frame, args=[message], daemon=True).start()
        elif "DATA" in message.message_type:
            if self.on_data is not None:
                threading.Thread(target=self.on_data, args=[message], daemon=True).start()


    def __questions__(self, question=socket_ops.Socket_Question):
        data = question.data
        message = messaging.translate_question(question, self)
        if 'IDENTIFY' in data:
            response = {"ip_data": self.ip_data, "mac": self.ip_data['mac']}

            if self.password == None and "VERIFY" in data:
                self.server.shutdown()
                # Your device set password to false, but the server asked for a password
                raise PermissionError("Connecting Server Error: Possible Spoof")

            if self.challengeType == verification.types.NONCE and self.password is not None and "VERIFY" in data:
                response["VERIFY"] = verification.solve_NONCE(data['VERIFY'], self.password)
                
            question.answer(response)

        elif "VIDEO" in message.message_type:
            if self.on_video_frame is not None:
                threading.Thread(target=self.on_video_frame, args=[message], daemon=True).start()
        elif "AUDIO" in message.message_type:
            if self.on_audio_frame is not None:
                threading.Thread(target=self.on_audio_frame, args=[message], daemon=True).start()
        elif "DATA" in message.message_type:
            if self.on_data is not None:
                threading.Thread(target=self.on_data, args=[message], daemon=True).start()
        elif "QUESTION" in message.message_type:
            if self.on_question is not None:
                threading.Thread(target=self.on_question, args=[message], daemon=True).start()

    def __disconnected__(self):
        self.connected = False
        if self.on_disconnect is not None:
            threading.Thread(target=self.on_disconnect, daemon=True).start()
        
    def connect(self, HOST=None, PORT=None):
        if HOST is not None:
            self.HOST = HOST
        if PORT is not None:
            self.PORT = PORT
        try:
            self.__connect_lock__ = True
            self.server.PORT = PORT
            self.server.HOST = HOST
            self.server.connect()
            while self.__connect_lock__:
                time.sleep(0.01)
            self.connected = True
        except Exception as e:
            print(e)
            self.__connect_lock__ = False
            raise e

    def disconnect(self):
        self.server.shutdown()




from essentials.network_ops import Get_GW, Get_IP
from essentials import GetPublicIP
from essentials.network_ops import IP_to_MAC, Get_IP_with_MAC


Get_Device_IPS = Get_IP
Device_IP_to_MAC = IP_to_MAC
Get_Gateway = Get_GW

def ignore(*args):
    pass


import netifaces
import socket
import time
import threading
from essentials import socket_ops_v2 as socket_ops
from essentials.network_ops import Get_GW, Get_IP

class Device(object):
    def __init__(self, ip, data):
        self.ip = ip
        self.discovery_data = data
    
    @property
    def json(self):
        data = {"ip": self.ip, "discovery_data": self.discovery_data}
        return data 

class Devices(object):
    def __init__(self):
        self.All = {}

    @property
    def json(self):
        data = {}
        for item in self.All:
            data[self.All[item].ip] = self.All[item].json
        return data

class Discovery_Server(object):
    def __init__(self, HOST=None, PORT=None, hotphrase="discovery", discovery_data="Server"):
        self.HOST = HOST
        self.PORT = PORT
        self.hotphrase = hotphrase
        self.discovery_data = discovery_data
        self.responder_function = self.__discovery_response__

        self.running = False

    def run(self, HOST=None, PORT=None):
        if self.running != False:
            return

        if HOST is not None:
            self.HOST = HOST
        if PORT is not None:
            self.PORT = PORT

        def __ignore__(_):
            pass

        self.server = socket_ops.UDP_Server(self.HOST, self.PORT, __ignore__, self.responder_function, max_buffer=len(self.hotphrase))
        self.running = True

    def shutdown(self):
        self.server.shutdown()
        self.running = False

    def __discovery_response__(self, data, connector=socket_ops.UDP_Server_Client):
        data = data.decode()
        if data == self.hotphrase:
            connector.send(self.discovery_data.encode())
        

class Discovery_Runner(object):

    def __init__(self, port, runners=5, hotphrase="discovery"):
        self.GW = Get_GW()
        self.IP = Get_IP()
        self.hotphrase = hotphrase
        self.runners = runners
        self.timeout = 5
        self.check_port = port
        self.base = ".".join(self.GW.split(".")[:3]) + "."
        self.running = 0
        self.Devices = Devices()

    def Collect(self, timeout=5, runners=None, hotphrase=None):
        if runners is not None:
            self.runners = runners
        if hotphrase is not None:
            self.hotphrase = hotphrase
        self.Devices = Devices()
        self.counted = 0
        self.timeout = timeout
        self.responses = 0

        runners = self.runners

        start = 1
        end = 0
        for i in range(1, runners + 1):
            end = 255//runners*i
            if i == runners:
                end = 255
            
            threading.Thread(target=self.__runner__, args=[start, end, self.check_port], daemon=True).start()
            self.running += 1
            
            start += (255//runners)

        while self.running > 0:
            print("[ DDS ] - Device Discovery Scan. Addresses Contacted:", self.counted, end="\r")
            time.sleep(0.01)
        print("[ DDS ] - Device Discovery Scan. Addresses Contacted:", self.counted)

        to = 0
        while to < timeout:
            print("[ DDS ] - Device Discovery Scan. Responses:", self.responses, end="\r")
            to += 0.01
            time.sleep(0.01)
        print("[ DDS ] - Device Discovery Scan. Responses:", self.responses)

        return self.Devices

    def __runner_data__(self, data, address):
        self.responses += 1
        self.Devices.All[address[0]] = Device(address[0], data.decode())
        

    def __runner__(self, start, end, port):
        while start <= end:
            try:
                if start == 255:
                    break
                rmIP = self.base + str(start)
                self.counted += 1

                try:
                    connector = socket_ops.UDP_Connector(rmIP, self.check_port, self.__runner_data__, 5)
                    connector.send(self.hotphrase.encode())                    

                except Exception as e:
                    print(e)
                    pass


            except KeyboardInterrupt:
                print("[ UKI ] - User Keyboard Interupt")
                exit()
            except TimeoutError:
                pass
            except Exception as e:
                #print(e)
                pass
            start += 1
        self.running -= 1


from essentials import socket_ops_v2 as socket_ops
import threading
from .. import Get_Gateway

class Broadcast_Receiver_Server(object):
    def __init__(self, HOST=None, PORT=None, on_data=None, max_buffer=1024):
        self.HOST = HOST
        self.PORT = PORT
        self.max_buffer=1024
        self.on_data = on_data

        self.running = False

    def run(self, HOST=None, PORT=None):
        if self.running != False:
            return

        if HOST is not None:
            self.HOST = HOST
        if PORT is not None:
            self.PORT = PORT

        def __ignore__(_):
            pass

        self.server = socket_ops.UDP_Server(self.HOST, self.PORT, __ignore__, self.__discovery_response__, max_buffer=self.max_buffer)
        self.running = True

    def shutdown(self):
        self.server.shutdown()
        self.running = False

    def __discovery_response__(self, data, connector=socket_ops.UDP_Server_Client):
        data = data.decode()
        if self.on_data is not None:
            threading.Thread(target=self.on_data, args=[data, connector], daemon=True).start()



def broadcast_message(message, port, on_response=None, boardcast_ip=None, max_buffer=1024):
    if boardcast_ip == None:
        boardcast_ip = Get_All_IP_Stat()['ext'][0]['broadcast']

    connector = socket_ops.UDP_Connector(boardcast_ip, port, on_response, 5, max_buffer=max_buffer)
    try:
        connector.send(message.encode())
    except:
        connector.send(message)


import cv2


def compress_cv2_image(img, quality=50):
    return cv2.imencode('.jpg', img, [cv2.IMWRITE_JPEG_QUALITY, quality])

def decompress_cv2_image(compressed_cv2):
    return cv2.imdecode(compressed_cv2, -1)


from essentials import socket_ops_v2 as socket_ops
import threading, time
from .. import verification
from . import messaging


class IOT_Server(object):
    
    def __init__(self, HOST=None, PORT=None, on_verified_new_connection_def=None, password=None, on_data=None, on_question=None, on_video_frame=None, on_audio_frame=None, on_disconnect=None, receive_confirm=True, receive_confirm_timeout=5, challenge_type=verification.types.NONCE, challenge_timeout=5, discovery_server=None):
        self.HOST = HOST
        self.PORT = PORT
        self.password = password
        self.challengeType = challenge_type
        self.on_verified_new_connection_def = on_verified_new_connection_def
        self.challenge_timeout = challenge_timeout
        self.discovery_server = discovery_server

        self.devices = {}

        self.on_data = on_data
        self.on_question = on_question
        self.on_video_frame = on_video_frame
        self.on_audio_frame = on_audio_frame
        self.on_disconnect = on_disconnect
        self.receive_confirm = receive_confirm
        self.receive_confirm_timeout = receive_confirm_timeout

    def run_server(self, HOST=None, PORT=None, Run_Discovery=True):
        if HOST is not None:
            self.HOST = HOST
        if PORT is not None:
            self.PORT = PORT

        if self.HOST is None or self.PORT is None:
            raise ValueError("HOST and PORT must be set to run a Server")

        self.server = socket_ops.Socket_Server_Host(self.HOST, self.PORT, self.on_connection, None, None, None, PYTHONIC_only=True)

        if self.discovery_server is not None:
            if self.discovery_server.PORT == self.PORT and self.discovery_server.HOST == self.HOST:
                raise ValueError("Clashing Server PORTS: Cannot have discovery port match Server PORT")
            if self.discovery_server.PORT is None:
                raise ValueError("Discovery port cannot be none")
            self.discovery_server.run()


    def on_connection(self, connector=socket_ops.Socket_Server_Client):
        connector.meta['validated'] = False
        print("new connection")
        identify_message = {"IDENTIFY": ["ip_data"]}
        if self.password != None:
            if self.challengeType == verification.types.NONCE:
                challenge = verification.create_NONCE(self.password)
            connector.meta['challenge'] = challenge
            identify_message['VERIFY'] = challenge['challenge']
            

        response = connector.ask(identify_message)

        if self.password is not None and 'VERIFY' in identify_message:
            try:
                if response['VERIFY'] != challenge['verify']:
                    raise PermissionError("error on verification")
            except:
                connector.send("error on verification")
                connector.shutdown()
                return

        try:
            connector.meta['ip_data'] = response['ip_data']
            try:
                connector.meta['mac'] = response['ip_data']['mac']
            except:
                connector.meta['mac'] = None
        except:
            connector.meta['ip_data'] = None
            connector.meta['mac'] = None

        iot_device = __IOT_Device__(
            self,
            connector,
            connector.meta['mac'],
            on_data=self.on_data,
            on_question=self.on_question,
            on_video_frame=self.on_video_frame,
            on_audio_frame=self.on_audio_frame,
            on_disconnect=self.on_disconnect,
            receive_confirm=self.receive_confirm,
            receive_confirm_timeout=self.receive_confirm_timeout
        )

        connector.meta['validated'] = verification.TimeStamp()

        if connector.meta['mac'] is not None:
            self.devices[connector.meta['mac']] = iot_device
        else:
            self.devices[connector.conID] = iot_device

        if self.on_verified_new_connection_def is not None:
            threading.Thread(target=self.on_verified_new_connection_def, args=[iot_device], daemon=True).start()
            
        connector.send("__connect_lock__ disable")    
        
    def __lost_connection__(self, connector):
        if self.on_disconnect is not None:
            threading.Thread(target=self.on_disconnect, args=[self.devices[connector.meta['mac']]], daemon=True).start()
        del self.devices[connector.meta['mac']]
    
    def ping_all(self):
        resp = {}
        for mac in self.devices:
            msg = self.devices[mac].ping()
            resp[mac] = msg.timing.round_trip
        return resp

    def broadcast_data(self, data):
        for mac in self.devices:
            self.devices[mac].send_data(data)

    def shutdown(self):
        self.server.Shutdown()

    def run_loop(self):
        while True:
            time.sleep(4)


class __IOT_Device__(object):

    def __init__(self, server, connector=socket_ops.Socket_Server_Client, mac=None, on_data=None, on_question=None, on_video_frame=None, on_audio_frame=None, on_disconnect=None, receive_confirm=True, receive_confirm_timeout=5, meta=None):
        self.server = server
        
        self.connector = connector
        self.connector.on_question = self.__questions__
        self.connector.on_data = self.__data__
        self.data_usage = self.connector.data_usage

        self.mac = mac
        self.on_data = on_data
        self.on_question = on_question
        self.on_video_frame = on_video_frame
        self.on_audio_frame = on_audio_frame
        self.on_disconnect = on_disconnect
        self.receive_confirm = receive_confirm
        self.receive_confirm_timeout = receive_confirm_timeout

        self.meta = meta


    def disconnect(self):
        self.connector.shutdown()


    def send_video_frame(self, frame, wait_till_confirm=False):
        msg = messaging.message(self.connector, frame, self.receive_confirm, self.receive_confirm_timeout, messaging.types.VIDEO, send_as_daemon= not wait_till_confirm)
        return msg

    def send_audio_frame(self, frame, wait_till_confirm=False):
        msg = messaging.message(self.connector, frame, self.receive_confirm, self.receive_confirm_timeout, messaging.types.AUDIO, send_as_daemon= not wait_till_confirm)
        return msg

    def send_data(self, data, wait_till_confirm=False):
        msg = messaging.message(self.connector, data, self.receive_confirm, self.receive_confirm_timeout, messaging.types.DATA, send_as_daemon= not wait_till_confirm)
        return msg

    def ping(self, wait_till_confirm=True):
        msg = messaging.message(self.connector, "PING", self.receive_confirm, self.receive_confirm_timeout, messaging.types.PING, send_as_daemon=not wait_till_confirm)
        return msg

    def ask(self, data, timeout=5):
        msg = messaging.message(self.connector, data, True, timeout, messaging.types.QUESTION)
        return msg

    def __data__(self, data, connector=socket_ops.Socket_Server_Client):
        if connector.meta['validated'] == False and self.password is not None:
            connector.shutdown()
            return

        try:
            connector = self.server.devices[connector.meta['mac']]
        except:
            connector = self.server.devices[connector.conID]
        
        message = messaging.translate_data(data, connector)

        if "VIDEO" == message.message_type:
            if self.on_video_frame is not None:
                threading.Thread(target=self.on_video_frame, args=[message], daemon=True).start()
        elif "AUDIO" in message.message_type:
            if self.on_audio_frame is not None:
                threading.Thread(target=self.on_audio_frame, args=[message], daemon=True).start()
        elif "DATA" in message.message_type:
            if self.on_data is not None:
                threading.Thread(target=self.on_data, args=[message], daemon=True).start()


    def __questions__(self, question=socket_ops.Socket_Question):
        connector = question.questioner

        if connector.meta['validated'] == False:
            connector.shutdown()
            return

        data = question.data
        message = messaging.translate_question(question, self)

        if "VIDEO" == message.message_type:
            if self.on_video_frame is not None:
                threading.Thread(target=self.on_video_frame, args=[message], daemon=True).start()
        elif "AUDIO" == message.message_type:
            if self.on_audio_frame is not None:
                threading.Thread(target=self.on_audio_frame, args=[message], daemon=True).start()
        elif "DATA" == message.message_type:
            if self.on_data is not None:
                threading.Thread(target=self.on_data, args=[message], daemon=True).start()
        elif "QUESTION" == message.message_type:
            if self.on_question is not None:
                threading.Thread(target=self.on_question, args=[message], daemon=True).start()
        elif "META_PUSH" == message.message_type:
            connector.meta['meta'] = message.data




from essentials import socket_ops_v2 as socket_ops
import threading
from essentials import TimeStamp
from .iot_device import IOT_Device


class directions(object):
    OUT = "OUT"
    IN = "IN"

class types(object):
    VIDEO = "VIDEO"
    AUDIO = "AUDIO"
    DATA = "DATA"
    PING = "PING"
    META_PUSH = "META_PUSH"
    QUESTION = "QUESTION"
    IDENTIFY = "IDENTIFY"


def get_msg_type(data):
    for item in "VIDEO.AUDIO.DATA.PING.IDENTIFY.QUESTION.META_PUSH".split("."):
        if item in data:
            return item

class Time_Keeper(object):

    def __init__(self):
        self.sent = TimeStamp()
    
    def confirmed(self, rcv_ts):
        self.received = rcv_ts
        self.round_trip = (TimeStamp() - self.sent) * 1000

def translate_question(question, connector):
    msg_type = get_msg_type(question.data)
    msg = message(connector, question.data[msg_type], message_direction=directions.IN, message_type=msg_type, question=question)
    if msg_type != types.QUESTION and msg_type != types.IDENTIFY:
        msg.__mark_recv__()
    return msg

def translate_data(data, connector):
    msg_type = get_msg_type(data)
    msg = message(connector, data[msg_type], message_direction=directions.IN, message_type=msg_type)
    return msg

class message(object):

    def __init__(self, connector=IOT_Device, data=None, recv_confirm=None, recv_timeout=5, message_type=types.DATA, message_direction=directions.OUT, send_as_daemon=True, question=None):
        self.connector = connector
        self.data = data
        self.recv_confirm = recv_confirm
        self.recv_timeout = recv_timeout
        self.message_type = message_type
        self.status = "pending"
        self.sent = False
        self.response = None
        self.error = False
        self.received = False
        self.timing = Time_Keeper()
        self.question = question
        

        if message_direction == directions.OUT:
            if self.message_type == types.QUESTION:
                self.__send__()
            elif recv_confirm == True:
                if send_as_daemon:
                    threading.Thread(target=self.__send__, daemon=True).start()
                else:
                    self.__send__()
            else:
                try:
                    self.connector.send({self.message_type: self.data})
                    self.status = "sent"
                    self.sent = True
                except Exception as x:
                    self.error = x
        else:
            self.status = "accepted"
            self.received = True

    def answer(self, data):
        self.question.answer({"ts": TimeStamp(), "resp": data})

    def __mark_recv__(self):
        self.question.answer({'ts': TimeStamp()})
        

    def __send__(self):
        try:
            self.status = "sending"
            resp = self.connector.ask({self.message_type: self.data}, self.recv_timeout)
            self.timing.confirmed(resp['ts'])
            try:
                self.response = resp['resp']
            except:
                self.response = None
            self.sent = True
            self.status = "received"
            self.received = True
        except Exception as x:
            self.error = x
            self.status = str(x)



import netifaces
import socket
import time
import threading
from essentials import socket_ops_v2 as socket_ops
from essentials.network_ops import Get_GW, Get_IP

class Device(object):
    def __init__(self, ip, data):
        self.ip = ip
        self.discovery_data = data
    
    @property
    def json(self):
        data = {"ip": self.ip, "discovery_data": self.discovery_data}
        return data 

class Devices(object):
    def __init__(self):
        self.All = {}

    @property
    def json(self):
        data = {}
        for item in self.All:
            data[self.All[item].ip] = self.All[item].json
        return data

class Discovery_Server(object):
    def __init__(self, HOST=None, PORT=None, hotphrase="discovery", discovery_data=False):
        self.HOST = HOST
        self.PORT = PORT
        self.hotphrase = hotphrase
        self.discovery_data = discovery_data

        self.running = False

    def run(self, HOST=None, PORT=None):
        if HOST is not None:
            self.HOST = HOST
        if PORT is not None:
            self.PORT = PORT

        def __ignore__(_):
            pass

        self.server = socket_ops.Socket_Server_Host(HOST, PORT, __ignore__, None, on_question=self.__discovery_response__, PYTHONIC_only=True)

    def __discovery_response__(self, question=socket_ops.Socket_Question):
        if question.data == self.hotphrase:
            question.answer(self.discovery_data)
        question.questioner.shutdown()

class Discovery_Runner(object):

    def __init__(self, port, rippers=5, hotphrase="discovery"):
        self.GW = Get_GW()
        self.IP = Get_IP()
        self.rippers = rippers
        self.hotphrase = hotphrase
        self.timeout = 5
        self.check_port = port
        self.base = ".".join(self.GW.split(".")[:3]) + "."
        self.running = 0
        self.Devices = Devices()

    def Collect(self, timeout=5, rippers=None, hotphrase=None):
        if rippers is not None:
            self.rippers = rippers
        if hotphrase is not None:
            self.hotphrase = hotphrase
        self.Devices = Devices()
        self.counted = 0
        self.timeout = timeout

        runners = rippers

        start = 1
        end = 0
        for i in range(1, runners + 1):
            end = 255//runners*i
            if i == runners:
                end = 255
            
            threading.Thread(target=self.__ripper__, args=[start, end, self.check_port], daemon=True).start()
            self.running += 1
            
            start += (255//runners)

        while self.running > 0:
            print("[ DDS ] - Device Discovery Scan. Addresses Contacted:", self.counted, end="\r")
            time.sleep(0.01)
        print("[ DDS ] - Device Discovery Scan. Addresses Contacted:", self.counted)

        return self.Devices
        

    def __ripper__(self, start, end, port):
        while start <= end:
            try:
                rmIP = self.base + str(start)
                self.counted += 1

                try:
                    def __ingnore__(*args):
                        pass

                    config = socket_ops.Configuration()
                    config.server_PYTHONIC_only = True
                    config.on_connection_close = __ingnore__
                    server = socket_ops.Socket_Connector(rmIP, port, config)
                    server.connect(self.timeout)
                    data = server.ask(self.hotphrase)
                    server.shutdown()
                    self.Devices.All[rmIP] = Device(rmIP, data)
                except:
                    pass


            except KeyboardInterrupt:
                print("[ UKI ] - User Keyboard Interupt")
                exit()
            except TimeoutError:
                pass
            except Exception as e:
                #print(e)
                pass
            start += 1
        self.running -= 1


from essentials import tokening, TimeStamp
from hashlib import md5

class types(object):
    NONCE = "nonce"



def create_NONCE(password):
    challenge = {
        "challenge": {
            "nonce": tokening.CreateToken(40),
            "ts": str(TimeStamp())
        }            
    }

    challenge['verify'] = md5((str(challenge['challenge']['ts']) + ":" + password + ":" + challenge['challenge']['nonce']).encode()).hexdigest()

    return challenge

def solve_NONCE(challenge, password):
    return md5((challenge['ts'] + ":" + password + ":" + challenge['nonce']).encode()).hexdigest()


from . import iot_device
import threading



class multi_outbound_connector(object):

    def __init__(self, on_connected_global=None, on_disconnected_global=None):
        self.on_connected_global = on_connected_global
        self.on_disconnected_global = on_disconnected_global
        self.outbound_connections = {}

    
    def connect_to_outbound_server(self, HOST, PORT, on_connected=None, on_disconnected=None):
        def disconnect():
            if on_disconnected is not None:
                threading.Thread(target=on_disconnected, daemon=True).start()
            if self.on_disconnected_global is not None:
                threading.Thread(target=on_disconnected, daemon=True).start()

        def connect():
            if on_connected is not None:
                threading.Thread(target=on_connected, daemon=True).start()
            if self.on_connected_global is not None:
                threading.Thread(target=self.on_connected_global, daemon=True).start()

        self.outbound_connections[HOST + ":" + str(PORT)] = iot_device.IOT_Device(HOST, PORT, connect, on_disconnect=disconnect)

        return self.outbound_connections[HOST + ":" + str(PORT)]